import pema
import straxen
import wfsim

_input_dict = dict(
    event_rate=20,  # Don't make too large -> overlapping truth info
    chunk_size=5,  # keep large -> less overhead but takes more RAM
    n_chunk=10,
    tpc_radius=straxen.tpc_r,
    tpc_length=straxen.tpc_z,
    drift_field=10,  # kV/cm
    energy_range=[1, 10],  # keV
    nest_inst_types=wfsim.NestId.ER,
)


def test_rand_instructions():
    pema.rand_instructions(**_input_dict)



import strax
import straxen
import wfsim
import pema
import os

export, __all__ = strax.exporter()


@export
def pema_context(
        base_dir: str,
        fax_config: str,
        cmt_run_id_sim: str,
        config_update: dict = None,
        cmt_version='global_v8',
        raw_dir=None,
        data_dir=None,
        raw_types=None,
        **kwargs,
) -> strax.Context:
    """
    Central context for pema, allows to init from a config.
    :param base_dir: Where store instructions,
    :param fax_config: fax configuration file
    :param cmt_run_id_sim: run_id for CMT (see straxen.contexts.xenonnt_simulation)
    :param cmt_version: the global correction version applied to the data
    :param config_update: Setup the config of the context
    :param raw_dir: Where to store the low level datatypes
    :param data_dir: Where to store the high level datatypes
    :param raw_types: Low level datatypes, stored separately from
        high level datatypes

    :kwargs: any kwargs are directly passed to the context
    :return: context
    """
    if not os.path.exists(base_dir):
        raise FileNotFoundError(
            f'Cannot use {base_dir} as base_dir. It does not exist.')

    config = dict(detector='XENONnT',
                  check_raw_record_overlaps=False,
                  fax_config=fax_config,
                  cmt_run_id_sim=cmt_run_id_sim,
                  )

    if config_update is not None:
        if not isinstance(config_update, dict):
            raise ValueError(f'Invalid config update {config_update}')
        config = strax.combine_configs(config, config_update)
    context_options = dict(
        free_options= ('n_nveto_pmts', 'channel_map', 'n_mveto_pmts',
                       'gain_model_nv', 'gain_model_mv', 'cmt_run_id_sim'),
    )
    st = straxen.contexts.xenonnt_simulation(
        fax_config=config['fax_config'],
        cmt_run_id_sim=cmt_run_id_sim,
        cmt_version=cmt_version,
        **context_options,
        **kwargs,
    )
    st.set_config(config)

    # Setup the plugins for nT
    # st.register(wfsim.RawRecordsFromFaxNT)
    st.register_all(pema.match_plugins)
    st._plugin_class_registry['peaks'].save_when = strax.SaveWhen.ALWAYS

    if raw_types is None:
        raw_types = (wfsim.RawRecordsFromFaxNT.provides +
                     straxen.plugins.PulseProcessing.provides)

    # Setup the storage, don't trust any of the stuff we get from xenonnt_simulation
    st.storage = []

    if raw_dir is not None:
        st.storage += [strax.DataDirectory(
            raw_dir,
            take_only=raw_types)]
    if data_dir is not None:
        st.storage += [strax.DataDirectory(
            data_dir,
            exclude=raw_types
        )]
    if not st.storage or not len(st.storage):
        raise RuntimeError('No storage, provide raw_dir and/or data_dir')
    return st



import typing as ty
import matplotlib.pyplot as plt
import numpy as np
import pema
import strax
import straxen
from strax.utils import tqdm
from straxen.analyses.waveform_plot import time_and_samples


@straxen.mini_analysis(
    requires=('peaks', 'peak_basics'),
    default_time_selection='touching',
    warn_beyond_sec=60)
def plot_peaks(peaks,
               seconds_range,
               t_reference,
               include_info=None,
               show_largest=100,
               single_figure=True,
               figsize=(10, 4),
               xaxis=True,
               ):
    if single_figure:
        plt.figure(figsize=figsize)
    plt.axhline(0, c='k', alpha=0.2)

    peaks = peaks[np.argsort(-peaks['area'])[:show_largest]]
    peaks = strax.sort_by_time(peaks)

    for p in peaks:
        plot_peak(p,
                  t0=t_reference,
                  include_info=include_info,
                  color={0: 'gray', 1: 'b', 2: 'g'}[p['type']])

    if xaxis == 'since_start':
        seconds_range_xaxis(seconds_range, t0=seconds_range[0])
    elif xaxis:
        seconds_range_xaxis(seconds_range)
        plt.xlim(*seconds_range)

    plt.ylabel("Intensity [PE/ns]")
    if single_figure:
        plt.tight_layout()


def plot_peak(p, t0=None, center_time=True, include_info=None, **kwargs):
    x, y = time_and_samples(p, t0=t0)
    kwargs.setdefault('linewidth', 1)

    # Plot waveform
    plt.plot(x, y,
             drawstyle='steps-pre',
             **kwargs,
             )

    if 'linewidth' in kwargs:
        del kwargs['linewidth']

    kwargs['alpha'] = kwargs.get('alpha', 1) * 0.2
    plt.fill_between(x, 0, y, step='pre', linewidth=0, **kwargs)

    # Mark extent with thin black line
    plt.plot([x[0], x[-1]],
             [y.max(), y.max()],
             c='k',
             alpha=0.3,
             linewidth=1,
             )

    # Mark center time with thin black line
    if center_time:
        if t0 is None:
            t0 = p['time']
        ct = (p['center_time'] - t0) / int(1e9)
        plt.axvline(ct,
                    c='k',
                    alpha=0.4,
                    linewidth=1,
                    linestyle='--',
                    )
    if include_info:
        info_str = '\n'.join([f'{inf}: {p[inf]:.1f}'
                              for inf in include_info])
        plt.text(x[-1],
                 y.max(),
                 info_str,
                 fontsize='xx-small',
                 ha='left',
                 va='top',
                 alpha=0.8,
                 bbox=dict(boxstyle="round",
                           fc="w",
                           alpha=0.5,
                           )
                 )


def _plot_truth(data, start_end, t_range):
    plt.title('Instructions')
    for pk, pi in enumerate(
            range(*strax.touching_windows(data, start_end)[0])):
        tpeak = data[pi]
        hatch_cycle = ['/', '*', '+', '|']
        _t_range = tpeak[['time', 'endtime']]
        x = np.array(list(_t_range))
        y = tpeak['n_pe'] / np.diff(x)
        ct = tpeak['t_mean_photon']
        stype = tpeak['type']
        plt.gca()
        plt.fill_between(
            [
                x[0] / 1e9,
                ct / 1e9,
                x[-1] / 1e9,
            ],
            [0, 0, 0],
            [0, 2 * y[0], 0],
            color={1: 'blue',
                   2: 'green',
                   0: 'gray',
                   6: 'orange',
                   4: 'purple',
                   }[stype],
            label=f'Peak S{stype}. {tpeak["n_pe"]} PE',
            alpha=0.4,
            hatch=hatch_cycle[pk]
        )
        plt.ylabel('Intensity [PE/ns]')
    for t in t_range:
        axvline(t / 1e9, label=f't = {t}')

    plt.legend(loc='lower left', fontsize='x-small')


def _plot_peak(st_default,
               truth_vs_default,
               default_label,
               peak_i,
               t_range,
               xlim,
               run_id,
               label_x_axis=False,
               ):
    plt.title(default_label)

    if run_id is None:
        run_id = truth_vs_default[peak_i]['run_id']

    st_default.plot_peaks(run_id,
                          single_figure=False,
                          include_info=['area', 'rise_time', 'tight_coincidence'],
                          time_range=t_range,
                          xaxis=label_x_axis,
                          )
    for t in t_range:
        axvline(t / 1e9, label=t)
    if label_x_axis:
        seconds_range_xaxis(xlim)
        plt.xlim(*xlim)
    plt.text(0.05, 0.95,
             truth_vs_default[peak_i]['outcome'],
             transform=plt.gca().transAxes,
             ha='left',
             va='top',
             bbox=dict(boxstyle="round",
                       fc="w",
                       )
             )

    plt.text(0.05, 0.1,
             '\n'.join(f'{prop[:10]}: {truth_vs_default[peak_i][prop]:.1f}'
                       for prop in
                       ['rec_bias', 'acceptance_fraction']),
             transform=plt.gca().transAxes,
             fontsize='small',
             ha='left',
             va='bottom',
             bbox=dict(boxstyle="round", fc="w"),
             alpha=0.8,
             )


def compare_truth_and_outcome(
        st: strax.Context,
        data: np.ndarray,
        **kwargs
) -> None:
    """
    Compare the outcomes of the truth and the reconstructed peaks

    :param st: the context of the current master, to compare with
    :param data: the  data consistent with the default
        context, can be cut to select certain data
    :param match_fuzz: Extend loading peaks this many ns to allow for
        small shifts in reconstruction. Will extend the time range left
        and right
    :param plot_fuzz: Make the plot slightly larger with this many ns
        for readability
    :param max_peaks: max number of peaks to be shown. Set to  1 for
        plotting a singe peak.
    :param label: How to label the default reconstruction
    :param fig_dir: Where to save figures (if None, don't save)
    :param show: show the figures or not.
    :param randomize: randomly order peaks to get a random sample of
        <max_peaks> every time
    :param run_id: Optional argument in case run_id is not a field in
        the data.
    :param raw: include raw-records-trace
    :param pulse: plot raw-record traces.
    :return: None
    """
    if kwargs:
        kwargs['different_by'] = None
    compare_outcomes(st=st,
                     data=data,
                     st_alt=None,
                     data_alt=None,
                     **kwargs,
                     )


def compare_outcomes(st: strax.Context,
                     data: np.ndarray,
                     st_alt: ty.Optional[strax.Context] = None,
                     data_alt: ty.Optional[np.ndarray] = None,
                     match_fuzz: int = 500,
                     plot_fuzz: int = 500,
                     max_peaks: int = 10,
                     default_label: str = 'default',
                     custom_label: str = 'custom',
                     fig_dir: ty.Union[None, str] = None,
                     show: bool = True,
                     randomize: bool = True,
                     different_by: ty.Optional[ty.Union[bool, str]] = 'acceptance_fraction',
                     run_id: ty.Union[None, str] = None,
                     raw: bool = False,
                     pulse: bool = True,
                     ) -> None:
    """
    Compare the outcomes of two contexts with one another. In order to
    allow for selections, we need to pass the data as second and third
    argument respectively.

    :param st: the context of the current master, to compare
        with st_custom
    :param data: the  data consistent with the default
        context, can be cut to select certain data
    :param st_alt: context wherewith to compare st_default
    :param data_alt: the data with the custom context, should be
        same length as truth_vs_default
    :param match_fuzz: Extend loading peaks this many ns to allow for
        small shifts in reconstruction. Will extend the time range left
        and right
    :param plot_fuzz: Make the plot slightly larger with this many ns
        for readability
    :param max_peaks: max number of peaks to be shown. Set to  1 for
        plotting a singe peak.
    :param default_label: How to label the default reconstruction
    :param custom_label:How to label the custom reconstruction
    :param fig_dir: Where to save figures (if None, don't save)
    :param show: show the figures or not.
    :param randomize: randomly order peaks to get a random sample of
        <max_peaks> every time
    :param different_by: Field to filter waveforms by. Only show
        waveforms where this field is different in data. If False, plot
        any waveforms from the two data sets.
    :param run_id: Optional argument in case run_id is not a field in
        the data.
    :param raw: include raw-records-trace
    :param pulse: plot raw-record traces.
    :return: None
    """

    if (st_alt is None) != (data_alt is None):
        raise RuntimeError('Both st_alt and data_alt should be specified simultaneously')
    _plot_difference = st_alt is not None

    if _plot_difference:
        _check_args(data, data_alt, run_id)
        peaks_idx = _get_peak_idxs_from_args(data,
                                             randomize,
                                             data_alt,
                                             different_by)
    else:
        _check_args(data, None, run_id)
        peaks_idx = _get_peak_idxs_from_args(data, randomize)

    for peak_i in tqdm(peaks_idx[:max_peaks]):
        try:
            if 'run_id' in data.dtype.names:
                run_mask = data['run_id'] == data[peak_i]['run_id']
                run_id = data[peak_i]['run_id']
            else:
                run_mask = np.ones(len(data), dtype=np.bool_)
            t_range, start_end, xlim = _get_time_ranges(data,
                                                        peak_i,
                                                        match_fuzz,
                                                        plot_fuzz)

            axes = iter(_get_axes_for_compare_plot(
                2
                + int(_plot_difference)
                + int(raw)
                + int(pulse))
            )

            plt.sca(next(axes))
            _plot_truth(data[run_mask], start_end, t_range)

            if raw:
                plt.sca(next(axes))
                st.plot_records_matrix(run_id,
                                       raw=True,
                                       single_figure=False,
                                       time_range=t_range,
                                       time_selection='touching',
                                       )
                for t in t_range:
                    axvline(t / 1e9)

            if pulse:
                plt.sca(next(axes))
                rr_simple_plot(st, run_id, t_range)

            plt.sca(next(axes))
            _plot_peak(st,
                       data,
                       default_label,
                       peak_i,
                       t_range,
                       xlim,
                       run_id,
                       label_x_axis=not _plot_difference,
                       )

            if _plot_difference:
                plt.sca(next(axes))
                _plot_peak(st_alt,
                           data_alt,
                           custom_label,
                           peak_i,
                           t_range,
                           xlim,
                           run_id,
                           label_x_axis=True,
                           )

            _save_and_show('example_wf_diff', fig_dir, show, peak_i)
        except (ValueError, RuntimeError) as e:
            print(f'Error making {peak_i}: {type(e)}, {e}')
            plt.show()


def rr_simple_plot(st, run_id, t_range):
    """
    Plot some raw-record pulses within (touching) the t_range
    :param st:
    :param run_id:
    :param t_range:
    :param legend:
    :return:
    """
    cmap = plt.cm.twilight(np.arange(straxen.n_tpc_pmts))
    raw_records = st.get_array(run_id, 'raw_records',
                               progress_bar=False,
                               time_range=t_range,
                               time_selection='touching',
                               )
    raw_records = np.sort(raw_records, order='channel')
    plt.ylabel('ADC counts')
    for rr in raw_records:
        y = rr['data'][:rr['length']]
        time = np.arange(len(y)) * rr['dt'] + rr['time']
        ch = rr['channel']
        idx = rr['record_i']
        plt.plot(time / 1e9,
                 y,
                 label=f'ch{ch:03}: rec_{idx}',
                 c=cmap[ch]
                 )
    for t in t_range:
        axvline(t / 1e9)


def axvline(v, **kwargs):
    vline_color = next(plt.gca()._get_lines.prop_cycler)['color']
    plt.axvline(v, color=vline_color, **kwargs)


def _get_peak_idxs_from_args(data_1, randomize, data_2=None, different_by=None):
    if different_by is not None and different_by:
        assert data_2 is not None
        peaks_idx = np.where(data_1[different_by] != data_2[different_by])[0]
    else:
        peaks_idx = np.arange(len(data_1))
    if randomize:
        np.random.shuffle(peaks_idx)
    return peaks_idx


def _check_args(truth_vs_default, truth_vs_custom=None, run_id=None):
    if 'run_id' not in truth_vs_default.dtype.names and run_id is None:
        raise ValueError('Either need a run_id or data with a run_id field!')
    if truth_vs_custom is not None and len(truth_vs_custom) != len(truth_vs_default):
        raise ValueError('Got different lengths for truth_vs_custom and truth_vs_default')


def _get_axes_for_compare_plot(n_axis):
    assert n_axis in [2, 3, 4, 5]
    _, axes = plt.subplots(
        n_axis,
        1,
        figsize=(10 * (n_axis / 3), 10),
        sharex=True,
        gridspec_kw={'height_ratios': [0.5, 1, 1, 1][:n_axis]}
    )
    return axes


def _save_and_show(name, fig_dir, show, peak_i):
    if fig_dir:
        pema.save_canvas(f'{name}_{peak_i}', save_dir=fig_dir)
    if show:
        plt.show()


def _get_time_ranges(truth_vs_custom, peak_i, matching_fuzz, plot_fuzz):
    t_range = (truth_vs_custom[peak_i]['time'] - matching_fuzz,
               truth_vs_custom[peak_i]['endtime'] + matching_fuzz)
    start_end = np.zeros(1, dtype=strax.time_fields)
    start_end['time'] = t_range[0]
    start_end['endtime'] = t_range[1]

    xlim = (t_range[0] - plot_fuzz) / 1e9, (t_range[1] + plot_fuzz) / 1e9
    return t_range, start_end, xlim


def seconds_range_xaxis(seconds_range, t0=None):
    """Make a pretty time axis given seconds_range"""
    plt.xlim(*seconds_range)
    ax = plt.gca()
    # disable for now ax.ticklabel_format(useOffset=False)
    xticks = plt.xticks()[0]
    if not len(xticks):
        return

    # Format the labels
    # I am not very proud of this code...
    def chop(x):
        return np.floor(x).astype(np.int64)

    if t0 is None:
        xticks_ns = np.round(xticks * int(1e9)).astype(np.int64)
    else:
        xticks_ns = np.round((xticks - xticks[0]) * int(1e9)).astype(np.int64)
    sec = chop(xticks_ns // int(1e9))
    ms = chop((xticks_ns % int(1e9)) // int(1e6))
    us = chop((xticks_ns % int(1e6)) // int(1e3))
    samples = chop((xticks_ns % int(1e3)) // 10)

    labels = [str(sec[i]) for i in range(len(xticks))]
    print_ns = np.any(samples != samples[0])
    print_us = print_ns | np.any(us != us[0])
    print_ms = print_us | np.any(ms != ms[0])
    if print_ms and t0 is None:
        labels = [l + f'.{ms[i]:03}' for i, l in enumerate(labels)]
        if print_us:
            labels = [l + r' $\bf{' + f'{us[i]:03}' + '}$'
                      for i, l in enumerate(labels)]
            if print_ns:
                labels = [l + f' {samples[i]:02}0' for i, l in enumerate(labels)]
        plt.xticks(ticks=xticks, labels=labels, rotation=90)
    else:
        labels = list(chop((xticks_ns // 10) * 10))
        labels[-1] = ""
        plt.xticks(ticks=xticks, labels=labels, rotation=0)
    if t0 is None:
        plt.xlabel("Time since run start [sec]")
    else:
        plt.xlabel("Time [ns]")



"""Simple script of pema to load requested modules"""
import socket
import strax
import straxen
import wfsim
import os
import pandas as pd
import numba
import time
import numpy as np
import pema
import json
import nestpy
from tqdm.notebook import tqdm
# Use nice fancy tqdm for simulating
wfsim.core.tqdm = tqdm
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
import multihist
from scipy.stats import norm
from collections import defaultdict

straxen.print_versions(['strax', 'straxen', 'wfsim', 'nestpy', 'ntauxfiles', 'pema'])

# Change the plots to have style and fashion
# Updated
params = {'axes.grid': True, 'font.size': 12,
          'axes.titlesize': 16, 'axes.labelsize': 14, 'axes.linewidth': 2,
          'xtick.labelsize': 14, 'ytick.labelsize': 14,
          'ytick.major.size': 8, 'ytick.minor.size': 4,
          'xtick.major.width': 2, 'xtick.minor.width': 2,
          'ytick.major.width': 2, 'ytick.minor.width': 2,
          'xtick.direction': 'in', 'ytick.direction': 'in',
          'legend.fontsize': 14, 'figure.facecolor': 'w',
          'figure.figsize': (10, 6),
          'lines.linewidth': 2,
          'image.cmap': 'plasma'}
plt.rcParams.update(params)



import warnings

import strax
import numba
import numpy as np
import pema
import logging
import straxen
from .matching import INT_NAN

export, __all__ = strax.exporter()

log = logging.getLogger('Pema matching')


@export
class MatchPeaks(strax.OverlapWindowPlugin):
    """
    Match WFSim truth to the outcome peaks. To this end use the
        matching algorithm of pema. Assign a peak-id to both the truth
        and the reconstructed peaks to be able to match the two. Also
        define the outcome of the matching (see pema.matching for
        possible outcomes).
    """
    __version__ = '0.5.0'
    depends_on = ('truth', 'truth_id', 'peak_basics', 'peak_id')
    provides = 'truth_matched'
    data_kind = 'truth'

    truth_lookup_window = straxen.URLConfig(
        default=int(1e9),
        help='Look back and forth this many ns in the truth info',
    )

    def compute(self, truth, peaks):
        log.debug(f'Starting {self.__class__.__name__}')
        truth = pema.append_fields(truth, 'area', truth['raw_area'])

        # hack endtime
        log.warning(f'Patching endtime in the truth')
        truth['endtime'] = truth['t_last_photon'].copy()

        log.info('Starting matching')
        truth_vs_peak, peak_vs_truth = pema.match_peaks(truth, peaks)

        # copy to the result buffer
        res_truth = np.zeros(len(truth), dtype=self.dtype)
        for k in self.dtype.names:
            res_truth[k] = truth_vs_peak[k]

        return res_truth

    def get_window_size(self):
        return self.config['truth_lookup_window']

    def infer_dtype(self):
        dtype = strax.dtypes.time_fields + [
            ((f'Id of element in truth', 'id'), np.int64),
            ((f'Outcome of matching to peaks', 'outcome'), pema.matching.OUTCOME_DTYPE),
            ((f'Id of matching element in peaks', 'matched_to'), np.int64)
        ]
        return dtype


@export
class AcceptanceComputer(strax.Plugin):
    """
    Compute the acceptance of the matched peaks. This is done on the
    basis of arbitrary settings to allow better to disentangle
    possible scenarios that might be undesirable (like splitting
    an S2 into small S1 signals that could affect event
    reconstruction).
    """
    __version__ = '2.0.2'
    depends_on = ('truth', 'truth_matched', 'peak_basics', 'peak_id')
    provides = 'match_acceptance'
    data_kind = 'truth'

    keep_peak_fields = straxen.URLConfig(
        default=('area', 'range_50p_area', 'area_fraction_top', 'rise_time', 'tight_coincidence'),
        help='Add the reconstructed value of these variables',
    )
    penalty_s2_by = straxen.URLConfig(
        default=(('misid_as_s1', -1.), ('split_and_misid', -1.),),
        help='Add a penalty to the acceptance fraction if the peak has the '
             'outcome. Should be tuple of tuples where each tuple should '
             'have the format of (outcome, penalty_factor)',
    )
    min_s2_bias_rec = straxen.URLConfig(
        default=0.85,
        help='If the S2 fraction is greater or equal than this, consider a '
             'peak successfully found even if it is split or chopped.',
    )

    def compute(self, truth, peaks):
        res = np.zeros(len(truth), self.dtype)
        res['time'] = truth['time']
        res['endtime'] = strax.endtime(truth)
        res['is_found'] = truth['outcome'] == 'found'

        peak_idx = truth['matched_to']
        mask = peak_idx != INT_NAN
        if np.sum(mask):
            # need to get at least one peak for each, even if we are going to remove those later
            sel_from_peaks = peak_idx[mask]
            sel_peaks = peaks[get_idx(sel_from_peaks, peaks['id'], INT_NAN)]

            if len(sel_peaks) != len(sel_from_peaks):
                raise ValueError(f'Got {len(sel_peaks)} != {len(sel_from_peaks)}')
            not_match = sel_from_peaks != sel_peaks['id']
            if np.any(not_match):
                for i, t_i, p_i in zip(not_match, sel_from_peaks, sel_peaks['id']):
                    print(i, t_i, p_i)
                raise ValueError
            for k in self.keep_peak_fields:
                res[f'rec_{k}'][mask] = sel_peaks[k]

        res['rec_bias'] = res['rec_area'] / truth['raw_area']

        # S1 acceptance is simply is the peak found or not
        s1_mask = truth['type'] == 1
        res['acceptance_fraction'][s1_mask] = res['is_found'][s1_mask].astype(np.float64)

        # For the S2 acceptance we calculate an arbitrary acceptance
        # that takes into account penalty factors and that S2s may be
        # split (as long as their bias fraction is not too small).
        s2_mask = truth['type'] == 2
        s2_outcomes = truth['outcome'][s2_mask].copy()
        s2_acceptance = (res[s2_mask]['rec_bias'] > self.config['min_s2_bias_rec']).astype(
            np.float64)
        for outcome, penalty in self.config['penalty_s2_by']:
            s2_out_mask = s2_outcomes == outcome
            s2_acceptance[s2_out_mask] = penalty

        # now update the acceptance fraction in the results
        res['acceptance_fraction'][s2_mask] = s2_acceptance

        return res

    def infer_dtype(self):
        dtype = strax.dtypes.time_fields + [
            ((f'Is the peak tagged "found" in the reconstructed data',
              'is_found'), np.bool_),
            ((f'Acceptance of the peak can be negative for penalized reconstruction',
              'acceptance_fraction'),
             np.float64),
            ((f'Reconstruction bias 1 is perfect, 0.1 means incorrect',
              'rec_bias'),
             np.float64),
        ]
        for descr in self.deps['peak_basics'].dtype_for('peak_basics').descr:
            # Add peak fields
            field = descr[0][1]
            if field in self.keep_peak_fields:
                dtype += [((descr[0][0], f'rec_{field}'), descr[1])]
        return dtype

    def setup(self):
        assert 'area' in self.keep_peak_fields


class AcceptanceExtended(strax.MergeOnlyPlugin):
    """Merge the matched acceptance to the extended truth"""
    __version__ = '0.1.0'
    depends_on = ('match_acceptance', 'truth', 'truth_id', 'truth_matched')
    provides = 'match_acceptance_extended'
    data_kind = 'truth'
    save_when = strax.SaveWhen.TARGET

    def setup(self):
        warnings.warn(f'match_acceptance_extended is deprecated use truth_extended',
                      DeprecationWarning)
        super().setup()


@export
class TruthExtended(strax.MergeOnlyPlugin):
    """Merge the matched acceptance to the extended truth"""
    __version__ = '0.1.0'
    depends_on = ('match_acceptance', 'truth', 'truth_id', 'truth_matched')
    provides = 'truth_extended'
    data_kind = 'truth'
    save_when = strax.SaveWhen.TARGET


class MatchEvents(strax.OverlapWindowPlugin):
    """
    Match WFSim truth to the outcome peaks. To this end use the
        matching algorithm of pema. Assign a peak-id to both the truth
        and the reconstructed peaks to be able to match the two. Also
        define the outcome of the matching (see pema.matching for
        possible outcomes).
    """
    __version__ = '0.1.0'
    depends_on = ('truth', 'events')
    provides = 'truth_events'
    data_kind = 'truth_events'

    truth_lookup_window = straxen.URLConfig(
        default=int(1e9),
        help='Look back and forth this many ns in the truth info',
    )
    check_event_endtime = straxen.URLConfig(
        default=True,
        help='Check that all events have a non-zero duration.',
    )
    sim_id_field = straxen.URLConfig(
        default='event_number',
        help='Group the truth info by this field. Options: ["event_number", "g4id"]',
    )
    dtype = strax.dtypes.time_fields + [
        ((f'First event number in event datatype within the truth event', 'start_match'), np.int64),
        ((f'Last (inclusive!) event number in event datatype within the truth event', 'end_match'),
         np.int64),
        ((f'Outcome of matching to events', 'outcome'), pema.matching.OUTCOME_DTYPE),
        ((f'Truth event number', 'truth_number'), np.int64),
    ]

    def compute(self, truth, events):
        unique_numbers = np.unique(truth[self.sim_id_field])
        res = np.zeros(len(unique_numbers), self.dtype)
        res['truth_number'] = unique_numbers
        fill_start_end(truth, res)
        if self.check_event_endtime:
            assert np.all(res['endtime'] > res['time'])
        assert np.all(np.diff(res['time']) > 0)

        tw = strax.touching_windows(events, res)
        tw_start = tw[:, 0]
        tw_end = tw[:, 1] - 1  # NB! This is now INCLUSIVE
        diff = np.diff(tw, axis=1)[:, 0]
        found = diff > 0

        # None unless found
        res['start_match'][~found] = pema.matching.INT_NAN
        res['end_match'][~found] = pema.matching.INT_NAN
        res['start_match'][found] = events[tw_start[found]]['event_number']
        res['end_match'][found] = events[tw_end[found]]['event_number']

        res['outcome'] = self.outcomes(diff)
        return res

    def get_window_size(self):
        return self.config['truth_lookup_window']

    @staticmethod
    def outcomes(diff):
        """Classify if the event_number"""
        outcome = np.empty(len(diff), dtype=pema.matching.OUTCOME_DTYPE)
        not_found_mask = diff < 1
        one_found_mask = diff == 1
        many_found_mask = diff > 1
        outcome[not_found_mask] = 'missed'
        outcome[one_found_mask] = 'found'
        outcome[many_found_mask] = 'split'
        return outcome


class PeakId(strax.Plugin):
    """Add id field to datakind"""
    depends_on = 'peak_basics'
    provides = 'peak_id'
    data_kind = 'peaks'
    __version__ = '0.0.0'

    peaks_seen = 0
    save_when = strax.SaveWhen.TARGET

    def infer_dtype(self):
        dtype = strax.time_fields
        id_field = [((f'Id of element in {self.data_kind}', 'id'), np.int64), ]
        return dtype + id_field

    def compute(self, peaks):
        res = np.zeros(len(peaks), dtype=self.dtype)
        res['time'] = peaks['time']
        res['endtime'] = peaks['endtime']
        peak_id = np.arange(len(peaks)) + self.peaks_seen
        res['id'] = peak_id
        self.peaks_seen += len(peaks)
        return res


class TruthId(PeakId):
    depends_on = 'truth'
    provides = 'truth_id'
    data_kind = 'truth'
    __version__ = '0.0.0'

    def compute(self, truth):
        assert_ordered_truth(truth)
        return super().compute(truth)


def fill_start_end(truth, truth_event, end_field='endtime'):
    """Set the 'time' and 'endtime' fields based on the truth"""
    truth_number = truth['event_number']
    starts = truth['time']
    stops = truth[end_field]
    _fill_start_end(truth_number, stops, starts, truth_event)


@numba.njit()
def _fill_start_end(truth_number, stops, starts, truth_event):
    for i, ev_i in enumerate(truth_event['truth_number']):
        mask = truth_number == ev_i
        start = starts[mask].min()
        stop = stops[mask].max()
        truth_event['time'][i] = start
        truth_event['endtime'][i] = stop


def assert_ordered_truth(truth):
    assert np.all(np.diff(truth['time']) >= 0), "truth is not sorted!"


@numba.njit
def get_idx(search_item, in_list, not_found=-99999):
    """Get index in <in_list> where the value is <searc_value>

    Assumes that <in_list> is sorted!

    :returns:
        a list of length (search item) where each value refers to the item in <in_list>
    """
    result = np.ones(len(search_item), dtype=np.int64) * not_found
    look_from = 0
    for i, search in enumerate(search_item):
        for k, v in enumerate(in_list[look_from:]):
            if v == search:
                result[i] = look_from + k
                look_from += k
                break
    for r in result:
        if r == not_found:
            raise ValueError()
    return result



"""Helpers for doing processing on dali"""
import os
import json
import strax
import typing as ty
from immutabledict import immutabledict
import subprocess
from collections import defaultdict
import pandas
import shutil

job_script = """\
#!/bin/bash
#SBATCH --partition {partition}
#SBATCH --qos {qos}
#SBATCH --account=pi-lgrandi
#SBATCH --ntasks=1
#SBATCH --output={log_file}
#SBATCH --error={log_file}
#SBATCH --cpus-per-task=1
#SBATCH --mem-per-cpu={mem}
#SBATCH --time={max_hours}

echo Processing job started

# Activate your environment you want to use
{bash_activate}

echo Environment activated
{cmd}
echo Processing job ended
"""

_qos = {'kicp': 'xenon1t-kicp'}

def write_script(fn, script, **kwargs):
    with open(fn, mode='w') as f:
        f.write(script.format(**kwargs))
    mode = os.stat(fn).st_mode
    mode |= (mode & 0o444) >> 2  # copy R bits to X
    os.chmod(fn, mode)


def write_dict_to_json(path: str,
                       to_write: dict, ):
    assert path.endswith('json')
    json_options = dict(sort_keys=True, indent=4)
    with open(path, mode='w') as f:
        f.write(json.dumps(to_write, **json_options))


class ProcessRun:
    """Class that allows for bookkeeping of runs of simulations"""
    log_file = None
    script_file = None
    base_dir_requires = ('configs', 'logs', 'scripts')
    process = None

    def __init__(self,
                 st: strax.context,
                 run_id: ty.Union[str, tuple, list],
                 target: ty.Union[str, tuple],
                 config: ty.Union[dict, immutabledict, None] = None,
                 ):
        self.st = st.new_context()
        self.run_id = strax.to_str_tuple(run_id)
        self.target = strax.to_str_tuple(target)
        if config is None:
            config = {}
        self.config = config
        self.st.set_config(config)

        self.base_dir = self.extract_base_dir(st)
        for subdir in self.base_dir_requires:
            os.makedirs(os.path.join(self.base_dir, subdir), exist_ok=True)

    def __repr__(self):
        rep = f'ProcessRun {self.run_id} - {self.target}:\n{self.config}'
        rep += f'\nwrite to {self.log_file} from {self.script_file}'
        return rep

    def all_stored(self, show_key=False, return_bool=False):
        bool_stored = True
        res = defaultdict(list)
        for r in self.run_id:
            res['number'].append(r)
            for t in self.target:
                stored = self.st.is_stored(r, t)
                bool_stored &= stored
                res[t].append(stored)
                if show_key:
                    res[f'{t}-key'] = self.st.key_for(r, t)

        df = pandas.DataFrame(res)
        df.set_index('number')
        if return_bool:
            return bool_stored
        return df

    @staticmethod
    def extract_base_dir(st):
        """Extract a base dir from the context (either from self or storage)"""
        if hasattr(st, 'base_dir'):
            base_dir = st.base_dir
        else:
            base_dir = None
            for sf in st.storage:
                if hasattr(sf, 'path'):
                    base_dir = sf.path
                    break
        if base_dir is None:
            raise FileNotFoundError('Cannot continue, no basedir')
        return base_dir

    def _fmt(self, directory: str, name: str) -> str:
        return os.path.join(self.base_dir, directory, name)

    def make_cmd(self,
                 debug=True,
                 not_lazy=True,
                 ):
        """
        return_command = just return the command, don't do the actual file stuf
        """
        st = self.st
        tot_config = st.config.copy()
        if 'channel_map' in tot_config:
            # Not saveable to JSON
            del tot_config['channel_map']

        run_id = self.run_id[0]
        target = self.target[-1]
        this_key = st.key_for(run_id, target)
        if len(self.run_id) > 1:
            job_name = f'{run_id}_{self.run_id[-1]}_{this_key}'
        else:
            job_name = f'{this_key}'
        if len(self.target) > 1:
            job_name += '-'.join(self.target[1:])

        conf_file = self._fmt('configs', f'config_{job_name}.json')

        cmd = (f'pema_straxer {" ".join(self.run_id)} '
               f'--target {" ".join(self.target)} '
               f'--context pema_context '
               f'--init_from_json {conf_file} '
               )

        if len(st.storage) == 1:
            raw_dir = None
            data_dir = st.storage[0].path
        elif len(st.storage) == 2:
            raw_dir = st.storage[0].path
            data_dir = st.storage[1].path
        else:
            raise ValueError
        context_init = dict(fax_config=self.st.config.get('fax_config', 'NO_FAX_CONFIG!'),
                            cmt_run_id_sim=self.st.config.get('cmt_run_id_sim', '026000'),
                            base_dir=self.base_dir,
                            config_update=tot_config,
                            raw_dir=raw_dir,
                            data_dir=data_dir,
                            )
        if debug:
            cmd += ' --debug'
        if not_lazy:
            cmd += ' --notlazy'
        if not sum([st.is_stored(run_id, 'raw_records') for run_id in self.run_id]):
            cmd += ' --build_lowlevel --rechunk_rr'

        write_dict_to_json(conf_file, context_init)
        return cmd, job_name

    def exec_dali(self,
                  cmd,
                  job_name,
                  bash_activate,
                  mem=2000,
                  partition='xenon1t',
                  max_hours="04:00:00"
                  ):
        self.log_file = self._fmt('logs', f'{job_name}.log')
        self.script_file = self._fmt('scripts', f'{job_name}.sh')
        script = job_script.format(
            bash_activate=bash_activate,
            log_file=self.log_file,
            cmd=cmd,
            mem=mem,
            partition=partition,
            qos=_qos.get(partition, partition),
            max_hours=max_hours)
        write_script(self.script_file, script)
        cp = subprocess.run(f'sbatch {self.script_file}', shell=True, capture_output=True)
        self.job_id = cp.stdout
        return self.job_id

    def exec_local(self, cmd, job_name):
        self.log_file = self._fmt('logs', f'{job_name}.log')
        self.script_file = f'{cmd} > {self.log_file}'
        # subprocess.Popen(self.script_file.split(' '), shell=True)
        # cp = subprocess.run(self.script_file, shell=True, capture_output=True)
        # return cp
        cmd = cmd.replace('  ', ' ')
        log = open(self.log_file, 'a')
        p = subprocess.Popen(cmd.split(' '),
                             stdout=log,
                             stderr=log,
                             universal_newlines=True)
        log.close()
        self.process = p

    def read_log(self):
        if self.log_file is None:
            raise RuntimeError('No logfile')
        f = open(self.log_file, "r")
        lines = [l for l in f]
        f.close()
        return lines

    def job_finished(self):
        finished = False
        for line in self.read_log()[-10:]:
            if 'Error' in line:
                raise JobFailedError(line)
            if 'ended' in line:
                finished = True
        return finished

    def purge_below(self,
                    delete_below='peaklets',
                    exclude=('raw_records', 'records',)):
        """
        Delete all the data below the given peaklet
        :param delete_below: target where below all the data
        :param exclude:
        :return:
        """
        for run_id in self.run_id:
            exclude_extended = ()
            for p in exclude:
                exclude_extended += self.st._plugin_class_registry[p].provides
            purgable = tuple(self.st._get_plugins((delete_below,), self.run_id[0]).keys())
            for p in purgable:
                if p in exclude_extended:
                    continue
                for sf in self.st.storage:
                    key = self.st.key_for(run_id, p)
                    try:
                        sbe, path = sf.find(key)
                    except strax.DataNotAvailable:
                        continue
                    if sbe != 'FileSytemBackend':
                        continue
                    for _path in [path, path + '_temp']:
                        if not os.path.exists(_path):
                            continue
                        print(_path)
                        shutil.rmtree(_path)


class JobFailedError(ValueError):
    """If a script job failes, raise this error"""



__version__ = '0.6.4'

from .matching import *
from .match_plugins import *
from . import match_plugins
from .summary_plots import *
from .wfsim_utils import *
from .scripts import *
from .contexts import *
from .misc import *
from .compare_plots import *



import straxen
import pema
import wfsim
import time
import matplotlib.pyplot as plt
import strax
import tempfile
import os
import unittest
import shutil
import uuid
import numpy as np

straxen.print_versions(['strax', 'straxen', 'wfsim', 'nestpy', 'pema'])

run_id = '008000'


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
class TestStack(unittest.TestCase):
    """
    Test the entire chain, from simulation to plotting the results

    Note:
        Important to notice is that despite this being slightly bad practice,
        the tests are ordered. This allows for speadier testing since we need
        to have some data to work with. Since that takes a while, we have
        decided to re-use the data.
    """

    @classmethod
    def setUpClass(cls):
        cls.set_temporary_test_folder()
        cls.set_script()

    @classmethod
    def set_temporary_test_folder(cls):
        temp_folder = uuid.uuid4().hex
        cls.tempdir = os.path.join(tempfile.gettempdir(), temp_folder)
        os.mkdir(cls.tempdir)

    @classmethod
    def set_script(cls):
        if not straxen.utilix_is_configured():
            return
        # setting up instructions like this may take a while. You can set e.g.
        instructions = dict(
            event_rate=50,  # Don't make too large -> overlapping truth info
            chunk_size=1,  # keep large -> less overhead but takes more RAM
            n_chunk=2,
            tpc_radius=straxen.tpc_r,
            tpc_length=straxen.tpc_z,
            drift_field=200,  # kV/cm
            energy_range=[2, 10],  # keV
            nest_inst_types=wfsim.NestId.ER,
        )
        temp_dir = cls.tempdir
        instructions_csv = os.path.join(temp_dir, 'inst.csv')

        pema.inst_to_csv(
            instructions_csv,
            get_inst_from=pema.rand_instructions,
            **instructions)

        config_update = {
            "detector": 'XENONnT',
            "fax_file": os.path.abspath(instructions_csv),
            "cmt_run_id_sim": run_id,
        }

        print("Temporary directory is ", temp_dir)
        os.chdir(temp_dir)

        st = pema.pema_context(fax_config='fax_config_nt_sr0_v0.json',
                               cmt_run_id_sim=run_id,
                               base_dir=temp_dir,
                               raw_dir=temp_dir,
                               data_dir=temp_dir,
                               config_update=config_update,
                               )
        st.set_context_config(
            {'allow_shm': True,
             'allow_lazy': False,
             'timeout': 300,
             'max_messages': 10,
             'allow_rechunk': False,
             }
        )
        script_writer = pema.ProcessRun(st, run_id,
                                        ('raw_records',
                                         'records',
                                         'peaklets',
                                         'truth_matched',
                                         'match_acceptance_extended'
                                         ))
        cls.script = script_writer

    def test_first_run_execute(self):
        print(f'Start script - context hash {self.script.st._context_hash()}')
        cmd, name = self.script.make_cmd()
        for t in self.script.target:
            self.script.st.make(self.script.run_id[0], t)
        self.script.exec_local(cmd, name)

        print(f'Starting\n\t{cmd}')
        t0 = time.time()
        print(self.script.process.communicate())
        print(f'took {time.time() - t0:.2f}s')
        print(f'Script done- context hash {self.script.st._context_hash()}')
        time.sleep(3)

        print(f'Done')
        print(f'Stored: {self.script.all_stored(show_key=True)}')
        check_all_stored = self.script.all_stored(return_bool=True)
        check_log_file = os.path.exists(self.script.log_file)
        try:
            check_did_finish = self.script.job_finished()
        except pema.scripts.JobFailedError:
            check_did_finish = False
        if not (check_all_stored and check_log_file and check_did_finish):
            print(f'Failed: '
                  f'check_all_stored {check_all_stored}, '
                  f'check_log_file: {check_log_file}, '
                  f'check_did_finish: {check_did_finish}, '
                  )
            if check_log_file:
                print(self.script.read_log())
            raise pema.scripts.JobFailedError(f'Job did not finish')

    def test_first_run_plugins(self):
        self.script.purge_below('match_acceptance_extended')
        for t in strax.to_str_tuple(self.script.target):
            for r in strax.to_str_tuple(self.script.run_id):
                if (self.script.st.get_save_when(t) > strax.SaveWhen.NEVER):
                    self.script.st.make(r, t)
                    assert self.script.st.is_stored(r, t)

    def test_next_print(self):
        print(self.script)

    def test_next_make_dali_script(self):
        # Just make sure we write some file, we are not actually going
        # to run it
        self.script.exec_dali('ls', 'test_job', 'dummy_activate')
        assert os.path.exists(self.script.script_file)

    def test_later_compare(self):
        st = self.script.st
        st2 = st.new_context()
        for t in strax.to_str_tuple(self.script.target):
            print(run_id, t)
            st2.make(run_id, t)
        peaks_1 = st.get_array(run_id, 'match_acceptance_extended')
        peaks_2 = st2.get_array(run_id, 'match_acceptance_extended')

        assert 'rec_area' in peaks_1.dtype.names
        assert np.sum(peaks_1[peaks_1['outcome'] == 'missed']['rec_area']) == 0

        if 'run_id' not in peaks_1.dtype.names:
            peaks_1 = pema.append_fields(peaks_1, 'run_id', [run_id] * len(peaks_1))
            peaks_2 = pema.append_fields(peaks_2, 'run_id', [run_id] * len(peaks_2))
        pema.compare_truth_and_outcome(
            st, peaks_1,
            max_peaks=2,
            show=False,
            fig_dir=self.tempdir,
            raw=True
        )
        plt.clf()
        pema.compare_outcomes(st, peaks_1,
                              st2, peaks_2,
                              max_peaks=2,
                              show=False,
                              different_by=None,
                              fig_dir=self.tempdir,
                              raw=False,
                              )
        plt.clf()
        if len(peaks_1):
            pema.summary_plots.plot_peak_matching_histogram(
                peaks_1,
                'n_photon',
                bin_edges=[0, int(peaks_1['n_photon'].max())]
            )
            plt.clf()
            pema.summary_plots.acceptance_plot(
                peaks_1,
                'n_photon',
                bin_edges=[0, int(peaks_1['n_photon'].max())]
            )
            plt.clf()

    def test_later_make_ev_matched(self):
        st = self.script.st
        st.get_array(run_id, 'truth_events')

    def test_later_rec_bas(self):
        st = self.script.st
        st2 = st.new_context()
        peaks_1 = st.get_array(run_id, 'match_acceptance_extended')
        peaks_2 = st2.get_array(run_id, 'match_acceptance_extended')
        peaks_1 = peaks_1[peaks_1['rec_bias'] > 0]
        peaks_2 = peaks_2[peaks_2['rec_bias'] > 0]
        peaks_1_kwargs = dict(bins=10)
        if len(peaks_1):
            pema.summary_plots.rec_plot(peaks_1, **peaks_1_kwargs)
            pema.summary_plots.reconstruction_bias(peaks_1, s1_kwargs=peaks_1_kwargs)
            plt.clf()
        if len(peaks_1) and len(peaks_2):
            if not np.sum(peaks_1['type'] == 1):
                return
            pema.summary_plots.rec_diff(
                peaks_1,
                peaks_2,
                s1_kwargs=peaks_1_kwargs,
                s2_kwargs=peaks_1_kwargs,
            )
            plt.clf()

    def test_later_inst_plot(self):
        st = self.script.st
        peaks = st.get_array(run_id, ('peaks', 'peak_basics'))
        st.plot_peaks(run_id, time_within=peaks[0])
        plt.clf()
        st.plot_peaks(run_id, time_within=peaks[0], xaxis='since_start')
        pema.plot_peak(peaks[0], t0=None)
        plt.clf()

        st.plot_peaks(run_id, time_within=peaks[0], xaxis=False)
        pema.save_canvas('test_fig', os.path.join(self.tempdir, 'figs'), tight_layout=True)

    def test_later_alternative_configs(self):
        st = self.script.st
        st.set_config(dict(check_event_endtime=False))
        st.make(run_id, 'truth_events')

#        st.set_config(dict(sim_id_field='g4id'))
#        st.make(run_id, 'truth_events')

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)



"""
Utility to match peaks from results of different processor versions / processor and simulator
"""
import numpy as np
import strax
import numba
import logging
from numpy.lib import recfunctions

log = logging.getLogger('Pema matching')

export, __all__ = strax.exporter()

INT_NAN = -99999
OUTCOME_DTYPE = '<U32'


@export
def match_peaks(allpeaks1,
                allpeaks2,
                matching_fuzz=0,
                unknown_types=(0,)):
    """
    Perform peak matching between two numpy record arrays with fields:
        time, endtime (or dt and length), id, type, area
    If a peak is split into many fragments (e.g. two close peaks split
    into three peaks), the results are unreliable and depend on which
    peak set is peaks1 and which is peaks2.

    Returns (allpeaks1, allpeaks2), each with two extra fields:
    outcome, matched_to:
        outcome: Can be one of:
            found:  Peak was matched 1-1 between peaks1 and peaks2 (type agrees,
             no other peaks in range).
                    Note that area, widths, etc. can still be quite different!
            missed: Peak is not present in the other list
            misid_as_XX: Peak is present in the other list, but has type XX
            merged: Peak is merged with another peak in the other list, the new
                'super-peak' has the same type
            merged_to_XX: As above, but 'super-peak' has type XX
            split: Peak is split in the other list, but more than one fragment
                has the same type as the parent.
            chopped: As split, but one or several fragments are unclassified,
                exactly one has the correct type.
            split_and_unclassified: As split, but all fragments are unclassified
                in the other list.
            split_and_misid: As split, but at least one fragment has a different
                peak type.
        matched_to: id of matching in *peak* in the other list if outcome is found
            or misid_as_XX, INT_NAN otherwise.
    """
    # Check required fields
    for i, d in enumerate((allpeaks1, allpeaks2)):
        assert hasattr(d, 'dtype'), 'Cannot work with non-numpy arrays'
        m = ''
        for k in ('area', 'type', 'id'):
            if k not in d.dtype.names:
                m += f'Argument {i} misses field {k} required for matching \n'
        if m != '':
            raise ValueError(m)
    log.debug('Appending fields')
    # Append id, outcome and matched_to fields
    allpeaks1 = append_fields(
        allpeaks1,
        ('outcome', 'matched_to'),
        (np.array(['missed'] * len(allpeaks1), dtype=OUTCOME_DTYPE),
         INT_NAN * np.ones(len(allpeaks1), dtype=np.int64)),
        dtypes=(OUTCOME_DTYPE, np.int64),
    )
    allpeaks2 = append_fields(
        allpeaks2,
        ('outcome', 'matched_to'),
        (np.array(['missed'] * len(allpeaks2), dtype=OUTCOME_DTYPE),
         INT_NAN * np.ones(len(allpeaks2), dtype=np.int64)),
        dtypes=(OUTCOME_DTYPE, np.int64),
    )

    log.debug('Getting windows')
    windows = strax.touching_windows(allpeaks1, allpeaks2, window=matching_fuzz)
    deep_windows = np.empty((0, 2), dtype=(np.int64, np.int64))
    # Each of the windows projects to a set of peaks in allpeaks2
    # belonging to allpeaks1. We also need to go the reverse way, which
    # I'm calling deep_windows below.
    if len(windows):
        # The order matters!! We matched allpeaks1->allpeaks2 so we now should match allpeaks2->allpeaks1
        deep_windows = get_deepwindows(windows, allpeaks2, allpeaks1, matching_fuzz)
        log.debug(f'Got {len(deep_windows)} deep windows and {len(windows)} windows')

    if not len(windows):
        # patch for empty data
        deep_windows = np.array([[-1, -1]], dtype=(np.int64, np.int64))
    assert np.shape(np.shape(deep_windows))[0] == 2, (
        f'deep_windows shape is wrong {np.shape(deep_windows)}\n{deep_windows}')

    # make array for numba
    unknown_types = np.array(unknown_types)

    # Inner matching
    _match_peaks(allpeaks1, allpeaks2, windows, deep_windows, unknown_types)
    return allpeaks1, allpeaks2


@numba.jit(nopython=True, nogil=True, cache=True)
def _match_peaks(allpeaks1, allpeaks2, windows, deep_windows, unknown_types):
    """See match_peaks_strax where we do the functional matching here"""
    # Loop over left and right bounds for peaks 1 by matching to peaks 2
    for peaks_1_i, (l1, r1) in enumerate(windows):
        peaks_1 = allpeaks1[l1:r1]
        if l1 == r1:
            continue

        for p1_i, p1 in enumerate(peaks_1):
            # Matching the other way around using deep_windows
            l2, r2 = deep_windows[peaks_1_i]

            peaks_2 = allpeaks2[l2:r2]
            matching_peaks = peaks_2
            if len(matching_peaks) == 0:
                pass

            elif len(matching_peaks) == 1:
                # A unique match! Hurray!
                p2 = matching_peaks[0]
                p1['matched_to'] = p2['id']
                p2['matched_to'] = p1['id']
                # Do the types match?
                if p1['type'] == p2['type']:
                    p1['outcome'] = 'found'
                    p2['outcome'] = 'found'
                else:
                    if _in(p1['type'], unknown_types):
                        p2['outcome'] = 'unclassified'
                    else:
                        p2['outcome'] = 'misid_as_s' + str(p1['type'])
                    if _in(p2['type'], unknown_types):
                        p1['outcome'] = 'unclassified'
                    else:
                        p1['outcome'] = 'misid_as_s' + str(p2['type'])
                    # If the peaks are unknown in both sets, they will
                    # count as 'found'.
                matching_peaks[0] = p2
            else:
                # More than one peak overlaps p1
                handle_peak_merge(parent=p1,
                                  fragments=matching_peaks,
                                  unknown_types=unknown_types)

            # matching_peaks is a copy, not a view, so we have to copy
            # the results over to peaks_2 manually Sometimes I wish
            # python had references...
            for i_in_matching_peaks, i_in_peaks_2 in enumerate(range(l2, r2)):
                allpeaks2[i_in_peaks_2] = matching_peaks[i_in_matching_peaks]

        # Match in reverse to detect merged peaks >1 peaks in 1 may
        # claim to be matched to a peak in 2, in which case we should
        # correct the outcome...
        for p2_i, p2 in enumerate(peaks_2):
            selection = peaks_1['matched_to'] == p2['id']
            matching_peaks = peaks_1[selection]
            if len(matching_peaks) > 1:
                handle_peak_merge(parent=p2,
                                  fragments=matching_peaks,
                                  unknown_types=unknown_types)

            # matching_peaks is a copy, not a view, so we have to copy
            # the results over to peaks_1 manually Sometimes I wish
            # python had references...
            for i_in_matching_peaks, i_in_peaks_1 in enumerate(
                    np.where(selection)[0]):
                peaks_1[i_in_peaks_1] = matching_peaks[i_in_matching_peaks]


@numba.jit(nopython=True)
def handle_peak_merge(parent, fragments, unknown_types):
    found_types = fragments['type']
    is_ok = found_types == parent['type']
    is_unknown = _in1d(found_types, unknown_types)
    is_misclass = _combine_and_flip(is_ok, is_unknown)
    # We have to loop over the fragments to avoid making a copy
    for i in range(len(fragments)):
        if is_unknown[i] or is_misclass[i]:
            if _in(parent['type'], unknown_types):
                fragments[i]['outcome'] = 'merged_to_unknown'
            else:
                fragments[i]['outcome'] = 'merged_to_s' + str(parent['type'])
        else:
            fragments[i]['outcome'] = 'merged'
        # Link the fragments to the parent
        fragments[i]['matched_to'] = parent['id']
    if np.any(is_misclass):
        parent['outcome'] = 'split_and_misid'
    # All fragments are either ok or unknown. If more than one fragment
    # is given the same class as the parent peak, then call it "split".
    elif len(np.where(is_ok)[0]) > 1:
        parent['outcome'] = 'split'
    elif np.all(is_unknown):
        parent['outcome'] = 'split_and_unclassified'
    # If exactly one fragment out of > 1 fragments is correctly
    # classified, then call the parent chopped
    else:
        parent['outcome'] = 'chopped'
    # We can't link the parent to all fragments. Link to the largest one
    _max_idx = _argmax(fragments['area'])
    parent['matched_to'] = fragments[_max_idx]['id']


@numba.jit(cache=True)
def get_deepwindows(windows, peaks_a, peaks_b, matching_fuzz):
    """Get matching window of the matched peak versus the original peak"""
    n_windows = len(windows)
    _deep_windows = np.ones((n_windows, 2), dtype=np.int64) * -1
    return _get_deepwindows(windows, peaks_a, peaks_b, matching_fuzz, _deep_windows)


@numba.njit(nogil=True, cache=True)
def _get_deepwindows(windows, peaks_a, peaks_b, matching_fuzz, _deep_windows):
    # Calculate the endtimes once
    peak_a_endtimes = strax.endtime(peaks_a)
    peak_b_endtimes = strax.endtime(peaks_b)

    # If we previously started on an index, the next index will not be 
    # before this, if we start here, we save time.
    prev_start = 0
    for window_i, w in enumerate(windows):
        l1, r1 = w
        if r1 - l1:
            match = strax.processing.general._touching_windows(
                peaks_a['time'][prev_start:], peak_a_endtimes[prev_start:],
                peaks_b[l1:r1]['time'], peak_b_endtimes[l1:r1],
                window=matching_fuzz)
            if len(match):
                # We have skipped the first prev_start items, add here
                this_window = match[0] + prev_start
                _deep_windows[window_i] = this_window
                prev_start = max(match[0][0], prev_start)
            else:
                # No match
                pass
    return _deep_windows


@export
def append_fields(base, names, data, dtypes=None, fill_value=-1,
                  usemask=False,  # Different from recfunctions default
                  asrecarray=False):
    """Append fields to numpy structured array
    Does nothing if array already has fields with the same name.
    """
    if isinstance(names, (tuple, list)):
        not_yet_in_data = True ^ np.in1d(names, base.dtype.names)
        if dtypes is None:
            dtypes = [d.dtype for d in data]
        # Add multiple fields at once
        return recfunctions.append_fields(
            base,
            np.array(names)[not_yet_in_data].tolist(),
            np.array(data)[not_yet_in_data].tolist(),
            np.array(dtypes)[not_yet_in_data].tolist(),
            fill_value,
            usemask,
            asrecarray)
    else:
        # Add single field
        if names in base.dtype.names:
            return base
        else:
            return recfunctions.append_fields(
                base, names, data, dtypes, fill_value, usemask, asrecarray)


# --- Numba functions where numpy does not suffice ---
@numba.njit
def _in1d(arr1, arr2):
    """
    Copy np.in1d logic for numba
    Five times faster than numpy #ohyeah
    """
    res = np.zeros(len(arr1), dtype=np.bool_)
    for i1 in range(len(arr1)):
        for v2 in arr2:
            if arr1[i1] == v2:
                res[i1] = 1
                break
    return res


@numba.jit(nopython=True)
def _in(val, arr):
    """
    Check if val is in array
    1.5x faster than val in np.array
    """
    for a in arr:
        if val == a:
            return True
    return False


@numba.jit(nopython=True)
def _argmax(arr):
    """
    Get index of max argument (np.argmax)
    Slightly faster than np.argmax
    """
    m = INT_NAN
    i = INT_NAN
    leng = len(arr)
    for j in range(leng):
        if arr[j] > m:
            m = arr[j]
            i = j
    return i


@numba.njit
def _combine_and_flip(arr1, arr2):
    """Combine the flipped arrays"""
    return _bool_flip(arr1).astype(np.bool_) & _bool_flip(arr2.astype(np.bool_))


@numba.njit
def _bool_flip(arr):
    """Use True ^ array"""
    res = np.zeros(len(arr), dtype=np.bool_)
    for i, a in enumerate(arr):
        if not a or a == 0:
            res[i] = 1
    return res



import numpy as np
import pandas as pd
import wfsim
import straxen
import strax
import nestpy
import typing as ty
from strax.utils import tqdm
from warnings import warn

export, __all__ = strax.exporter()


@export
def rand_instructions(
        event_rate: int,
        chunk_size: int,
        n_chunk: int,
        drift_field: float,
        energy_range: ty.Union[tuple, list, np.ndarray],
        tpc_length: float = straxen.tpc_z,
        tpc_radius: float = straxen.tpc_r,
        nest_inst_types: ty.Union[ty.List[int], ty.Tuple[ty.List], np.ndarray, None] = None,
) -> dict:
    """
    Generate instructions to run WFSim
    :param event_rate: # events per second
    :param chunk_size: the size of each chunk
    :param n_chunk: the number of chunks
    :param energy_range: the energy range (in keV) of the recoil type
    :param tpc_length: the max depth of the detector
    :param tpc_radius: the max radius of the detector
    :param nest_inst_types: the
    :param drift_field:
    :return:
    """
    if nest_inst_types is None:
        nest_inst_types = [7]

    n_events = event_rate * chunk_size * n_chunk
    total_time = chunk_size * n_chunk

    inst = np.zeros(2 * n_events, dtype=wfsim.instruction_dtype)
    inst[:] = -1

    uniform_times = total_time * (np.arange(n_events) + 0.5) / n_events

    inst['time'] = np.repeat(uniform_times, 2) * int(1e9)
    inst['event_number'] = np.digitize(inst['time'],
                                       1e9 * np.arange(n_chunk) *
                                       chunk_size) - 1
    inst['type'] = np.tile([1, 2], n_events)

    r = np.sqrt(np.random.uniform(0, tpc_radius ** 2, n_events))
    t = np.random.uniform(-np.pi, np.pi, n_events)
    inst['x'] = np.repeat(r * np.cos(t), 2)
    inst['y'] = np.repeat(r * np.sin(t), 2)
    inst['z'] = np.repeat(np.random.uniform(-tpc_length, 0, n_events), 2)

    # Here we'll define our XENON-like detector
    nest_calc = nestpy.NESTcalc(nestpy.VDetector())
    nucleus_A = 131.293
    nucleus_Z = 54.
    lxe_density = 2.862  # g/cm^3   #SR1 Value

    energy = np.random.uniform(*energy_range, n_events)
    quanta = []
    exciton = []
    recoil = []
    e_dep = []
    for energy_deposit in tqdm(energy, desc='generating instructions from nest'):
        interaction_type = np.random.choice(nest_inst_types)
        interaction = nestpy.INTERACTION_TYPE(interaction_type)
        y = nest_calc.GetYields(interaction,
                                energy_deposit,
                                lxe_density,
                                drift_field,
                                nucleus_A,
                                nucleus_Z,
                                )
        q = nest_calc.GetQuanta(y, lxe_density)
        quanta.append(q.photons)
        quanta.append(q.electrons)
        exciton.append(q.excitons)
        exciton.append(0)
        # both S1 and S2
        recoil += [interaction_type, interaction_type]
        e_dep += [energy_deposit, energy_deposit]

    inst['amp'] = quanta
    inst['local_field'] = drift_field
    inst['n_excitons'] = exciton
    inst['recoil'] = recoil
    inst['e_dep'] = e_dep
    for field in inst.dtype.names:
        if np.any(inst[field] == -1):
            warn(f'{field} is not (fully) filled')
    return inst


@export
def inst_to_csv(csv_file: str,
                get_inst_from=rand_instructions,
                **kwargs):
    """
    Write instructions to csv file
    :param csv_file: path to the csv file
    :param get_inst_from: function to modify instructions and generate
    S1 S2 instructions from
    :param kwargs: key word arguments to give to the get_inst_from-function
    """
    df = pd.DataFrame(get_inst_from(**kwargs))
    if np.any(df['amp'] <= 0):
        warn('Removing zero amplitude from instruction, but that shouldn\'t be here')
        df = df[df['amp'] > 0]
    df.to_csv(csv_file, index=False)



import matplotlib.pyplot as plt
import os
import typing as ty

from strax import exporter as strax_exporter

export, __all__ = strax_exporter()


@export
def save_canvas(name: str,
                save_dir: str = './figures',
                dpi: int = 200,
                tight_layout: bool = False,
                pickle_dump: bool = False,
                other_formats: ty.Union[ty.List[str], None] = None,
                ) -> None:
    """
    Wrapper for saving current figure. Saves to PNG by default and also
    adds pdf and svg formats

    :param name: How to name the figure (no extension required)
    :param save_dir: Base where to store figures
    :param dpi: dots per inch for saving the option
    :param tight_layout: if true use plt.tight_layout
    :param pickle_dump: DEPRECATED option
    :param other_formats: other formats to save in (by default svg and pdf)
    :return: None
    """
    assert not pickle_dump, "Allowing pickle dumps is deprecated"
    if other_formats is None:
        other_formats = 'pdf svg'.split()
    if not os.path.exists(save_dir):
        os.makedirs(save_dir + '/.')
    for sub_folder in other_formats:
        sub_dir = os.path.join(save_dir, sub_folder)
        if not os.path.exists(sub_dir):
            os.makedirs(sub_dir)
    if tight_layout:
        plt.tight_layout()

    plt.savefig(f"{save_dir}/{name}.png", dpi=dpi, bbox_inches="tight")

    for extension in other_formats:
        plt.savefig(
            os.path.join(
                save_dir,
                extension,
                f'{name}.{extension}'),
            dpi=dpi,
            bbox_inches="tight")



from unittest import TestCase, skipIf
import numpy as np
import pema
import straxen


def test_import():
    from pema import matching
    from pema import MatchPeaks
    from pema import append_fields
    try:
        from pema import _plot_peak_matching_histogram
        raise ValueError(
            '_plot_peak_matching_histogram should not be available at top level')
    except (ImportError, ValueError, FileNotFoundError, ModuleNotFoundError):
        # Good we cannot import this (which is what we want)
        pass
    print('done')


class SimpleTests(TestCase):
    """Odd bunch of tests that can be removed if needed"""

    def check_raises_check_args(self):
        truth = np.zeros(1, dtype=[(('bla bla', 'bla'), np.int8)])
        with self.assertRaises(ValueError):
            pema.compare_plots._check_args(truth, run_id=None)
        with self.assertRaises(ValueError):
            pema.compare_plots._check_args([], truth_vs_custom=truth, run_id='bla')

    def test_show(self):
        pema.compare_plots._save_and_show('a', '.test', show=True, peak_i=1)

    def test_warnings_for_matching(self):
        with self.assertRaises(ValueError):
            peak1 = np.zeros(1, dtype=[(('bla bla', 'bla'), np.int8)])
            peak2 = peak1
            pema.match_peaks(peak1, peak2)

    @skipIf(not straxen.utilix_is_configured(), 'no utilix file')
    def test_context(self):
        with self.assertRaises(FileNotFoundError):
            pema.pema_context(base_dir='/the/moon/path', fax_config='bla', cmt_run_id_sim='1')
        with self.assertRaises(ValueError):
            pema.pema_context(base_dir='./', config_update=1, fax_config='bla', cmt_run_id_sim='1')
        with self.assertRaises(RuntimeError):
            pema.pema_context(base_dir='./',
                              config_update={},
                              fax_config='fax_config_nt_design.json',
                              cmt_run_id_sim='1')



import pema
import numpy as np
from hypothesis import given, strategies, example, settings
import strax


def _create_dummy_records(data_length, n_data_types, truth_length, n_truth_types, max_duration, ):
    _DTYPE = [(('Start time since unix epoch [ns]', 'time'), np.int64),
              (('Exclusive end time since unix epoch [ns]', 'endtime'), np.int64),
              (('Number', 'id'), np.int64),
              (('type of p', 'type'), np.int16),
              (('area of p', 'area'), np.float64)]

    data = np.zeros(data_length, dtype=_DTYPE)
    data['id'] = np.arange(data_length)
    data['time'] = np.random.randint(0, 1e9, data_length)
    data.sort(order='time')

    data['endtime'] = data['time']
    if max_duration:
        # randint does not work for 0, 0 interval, hence we only add it if max_duration>0
        data['endtime'] += np.random.randint(0, max_duration, data_length)
        data['endtime'][:-1] = np.clip(data['endtime'][:-1], data['time'][:-1], data['time'][1:], )
    data['type'] = np.random.randint(0, n_data_types + 1, data_length)

    truth = np.zeros(truth_length, dtype=_DTYPE)
    truth['time'] = np.random.randint(0, 1e9, truth_length)
    truth.sort(order='time')

    truth['endtime'] = truth['time']
    if max_duration:
        # randint does not work for 0, 0 interval, hence we only add it if max_duration>0
        truth['endtime'] += np.random.randint(0, max_duration, truth_length)
        truth['endtime'][:-1] = np.clip(truth['endtime'][:-1], truth['time'][:-1],
                                        truth['time'][1:])
    truth['type'] = np.random.randint(0, n_truth_types + 1, truth_length)
    return data, truth


@settings(max_examples=100, deadline=None)
@given(strategies.integers(min_value=0, max_value=100),
       strategies.integers(min_value=0, max_value=2), )
@example(length=3, dtypenum=0)
def test_combine_and_flip(length, dtypenum):
    """Check that combine and flip works correctly"""
    known_dtypes = (np.bool_, np.int8, np.int64)
    dtype = known_dtypes[dtypenum]
    arr1 = np.random.randint(0, 2, length).astype(dtype)
    arr2 = np.random.randint(0, 2, length).astype(dtype)
    res = (True ^ arr1) & (True ^ arr2)
    numba_res = pema.matching._combine_and_flip(arr1, arr2)
    assert np.all(res == numba_res)


@settings(max_examples=100, deadline=None)
@given(
    # Max length of the data
    strategies.integers(min_value=0, max_value=1000),
    # Max length of the truth
    strategies.integers(min_value=0, max_value=1000),
    # Max duration in [ns]
    strategies.integers(min_value=0, max_value=10_000),
    # Number of peak-types in data
    strategies.integers(min_value=1, max_value=10),
    # Number of peak-types in truth
    strategies.integers(min_value=1, max_value=10),
)
@example(data_length=0,
         truth_length=10,
         max_duration=10,
         n_data_types=2,
         n_truth_types=2)
def test_matching(data_length,
                  truth_length,
                  max_duration,
                  n_data_types,
                  n_truth_types):
    data, truth = _create_dummy_records(data_length,
                                        n_data_types,
                                        truth_length,
                                        n_truth_types,
                                        max_duration,
                                        )
    d_matched, t_matched = pema.match_peaks(data, truth)
    assert len(d_matched) == len(data)
    assert len(t_matched) == len(truth)


@settings(max_examples=100, deadline=None)
@given(
    # Max length of the data
    strategies.integers(min_value=0, max_value=1000),
    # Max length of the truth
    strategies.integers(min_value=0, max_value=1000),
    # Max duration in [ns]
    strategies.integers(min_value=0, max_value=10_000),
    # Number of peak-types in data
    strategies.integers(min_value=0, max_value=10),
    # Number of peak-types in truth
    strategies.integers(min_value=1, max_value=10),
    # Matching fuzz
    strategies.integers(min_value=0, max_value=10_000),
)
@example(data_length=0,
         truth_length=10,
         max_duration=10,
         n_data_types=2,
         n_truth_types=2,
         matching_fuzz=1)
def test_deepwindows(data_length,
                     truth_length,
                     max_duration,
                     n_data_types,
                     n_truth_types,
                     matching_fuzz):
    data, truth = _create_dummy_records(data_length,
                                        n_data_types,
                                        truth_length,
                                        n_truth_types,
                                        max_duration, )
    allpeaks1 = truth.copy()
    allpeaks2 = data.copy()

    windows = strax.touching_windows(allpeaks1, allpeaks2, window=matching_fuzz)
    deepwindows_simple = get_deepwindows(windows, allpeaks1, allpeaks2, matching_fuzz)
    deepwindows_numba = pema.matching.get_deepwindows(windows, allpeaks1, allpeaks2, matching_fuzz)
    if len(deepwindows_simple) > 0 and len(deepwindows_numba) > 0:
        assert np.all(deepwindows_simple == deepwindows_numba)
    assert len(deepwindows_simple) == len(deepwindows_numba)


def get_deepwindows(windows, peaks_a, peaks_b, matching_fuzz):
    """Do it the non-numba way, should work as well but is slower"""
    _deep_windows = []
    for l1, r1 in windows:
        this_window = [-1, -1]
        if r1 - l1:
            match = strax.touching_windows(peaks_a,
                                           peaks_b[l1:r1],
                                           window=matching_fuzz)
            if len(match) > 0:
                this_window = match[0]
            else:
                pass
        _deep_windows.append(this_window)
    deep_windows = np.array(_deep_windows, dtype=(np.int64, np.int64))
    return deep_windows


def test_peak_merges():
    """Test some examples"""
    dtype = [(('outcome of matching', 'outcome'), pema.matching.OUTCOME_DTYPE),
             (('Number', 'id'), np.int64),
             (('matched to', 'matched_to'), np.int64),
             (('type of p', 'type'), np.int16),
             (('area of p', 'area'), np.float64)]
    unknown_types = np.array([0])
    parent = np.zeros(1, dtype=dtype)
    fragment = np.zeros(2, dtype=dtype)

    parent['type'] = 1
    fragment['type'] = 1, 2
    pema.matching.handle_peak_merge(parent[0], fragment, unknown_types)
    assert parent['outcome'] == 'split_and_misid'
    assert fragment[0]['outcome'] == 'merged'
    assert fragment[1]['outcome'] == 'merged_to_s1'

    parent['type'] = 1
    fragment['type'] = 1, 0
    pema.matching.handle_peak_merge(parent[0], fragment, unknown_types)
    assert parent['outcome'] == 'chopped'

    parent['type'] = 1
    fragment['type'] = 1, 2
    pema.matching.handle_peak_merge(parent[0], fragment, unknown_types)
    assert parent['outcome'] == 'split_and_misid'

    parent['type'] = 1
    fragment['type'] = 1, 1
    pema.matching.handle_peak_merge(parent[0], fragment, unknown_types)
    assert parent['outcome'] == 'split'

    parent['type'] = 0
    fragment['type'] = 1, 1
    pema.matching.handle_peak_merge(parent[0], fragment, unknown_types)
    assert fragment[0]['outcome'] == 'merged_to_unknown'

    parent['type'] = 1
    fragment['type'] = 0, 0
    pema.matching.handle_peak_merge(parent[0], fragment, unknown_types)
    assert parent['outcome'] == 'split_and_unclassified'



import matplotlib.pyplot as plt
from scipy.stats import beta
from scipy import stats
import numpy as np
from multihist import Hist1d
import strax
import multihist
from scipy.stats import norm
from immutabledict import immutabledict
from copy import deepcopy

export, __all__ = strax.exporter()

outcome_colors = {
    'found': 'darkblue',
    'chopped': 'mediumslateblue',

    'missed': 'red',
    'merged': 'turquoise',
    'split': 'purple',

    'misid_as_s2': 'orange',
    'misid_as_s1': 'goldenrod',
    'split_and_misid': 'darkorange',
    'merged_to_s2': 'chocolate',
    'merged_to_s1': 'sandybrown',
    'merged_to_unknown': 'khaki',

    'unclassified': 'green',
    'split_and_unclassified': 'seagreen',
    'merged_and_unclassified': 'limegreen',
}

DEFAULT_NPH_LABEL = r'$\#\mathrm{Photon\ detected\ (excl. DPE)}$'


@export
def peak_matching_histogram(results, histogram_key, bin_edges):
    """
    Make 1D histogram of peak matching results (=peaks with extra fields)
    added by histogram_key
    """

    if histogram_key not in results.dtype.names:
        raise ValueError(
            'Histogram key %s should be one of the columns in results: %s' % (
                histogram_key,
                results.dtype.names))

    # How many true peaks do we have in each bin in total?
    n_peaks_hist = Hist1d(results[histogram_key], bin_edges)
    hists = {'_total': n_peaks_hist}

    for outcome in np.unique(results['outcome']):
        # Histogram the # of peaks that have this outcome
        hist = Hist1d(results[results['outcome'] == outcome][histogram_key],
                      bins=n_peaks_hist.bin_edges)
        hists[outcome] = hist

    return hists


@export
def plot_peak_matching_histogram(*args, **kwargs):
    hists = peak_matching_histogram(*args, **kwargs)
    _plot_peak_matching_histogram(hists)


def _plot_peak_matching_histogram(hists):
    """
    Make 1D histogram of peak matching results (=peaks with extra fields)
    added by histogram_key
    """

    n_peaks_hist = hists['_total']

    for outcome, hist in hists.items():
        hist = hist.histogram.astype(np.float)

        if outcome == '_total':
            continue

        print("\t%0.2f%% %s" % (100 * hist.sum() / n_peaks_hist.n, outcome))

        # Compute Errors on estimate of a proportion
        # Should have vectorized this... lazy
        # Man this code is ugly!!!!
        limits_d = []
        limits_u = []
        for i, x in enumerate(hist):
            limit_d, limit_u = binom_interval(x,
                                              total=n_peaks_hist.histogram[i])
            limits_d.append(limit_d)
            limits_u.append(limit_u)
        limits_d = np.array(limits_d)
        limits_u = np.array(limits_u)

        # Convert hist to proportion
        hist /= n_peaks_hist.histogram.astype('float')

        color = outcome_colors.get(outcome, np.random.rand(3, ))
        plt.errorbar(x=n_peaks_hist.bin_centers,
                     y=hist,
                     yerr=[hist - limits_d, limits_u - hist],
                     label=outcome,
                     color=color,
                     linestyle='-' if outcome == 'found' else '',
                     drawstyle='steps-mid',
                     marker='s')

        # Wald intervals: not so good
        # errors = np.sqrt(
        #     hist*(1-hist)/all_true_peaks_histogram
        # )
        # plt.errorbar(x=bin_centers, y=hist, yerr = errors, label=outcome)

    plt.xlim(n_peaks_hist.bin_edges[0], n_peaks_hist.bin_edges[-1])
    plt.ylabel('Fraction of peaks')
    plt.ylim(0, 1)
    plt.legend(loc='lower right', shadow=True)
    legend = plt.legend(loc='best', prop={'size': 10})
    if legend and legend.get_frame():
        legend.get_frame().set_alpha(0.8)


@export
def binom_interval(success, total, conf_level=0.95):
    """
    Confidence interval on binomial - using Jeffreys interval
    Code stolen from https://gist.github.com/paulgb/6627336
    Agrees with http://statpages.info/confint.html for binom_interval(1, 10)
    """
    # Should we add a special case for success = 0 or = total? see wikipedia
    quantile = (1 - conf_level) / 2.
    lower = beta.ppf(quantile, success, total - success + 1)
    upper = beta.ppf(1 - quantile, success + 1, total - success)
    # If something went wrong with a limit calculation, report the trivial limit
    if np.isnan(lower):
        lower = 0
    if np.isnan(upper):
        upper = 1
    return lower, upper


def get_interval(x, n, found):
    one_sigma = stats.norm.cdf(1) - stats.norm.cdf(-1)
    eff = found / n

    limits = np.zeros((len(x), 2))
    for i, w in enumerate(x):
        limits[i, :] = binom_interval(found[i], total=n[i], conf_level=one_sigma)
    yerr = np.abs(limits.T - eff)
    return eff, yerr


def calc_arb_acceptance(data, on_axis, bin_edges, nbins=None, ) -> tuple:
    """Calculate acceptance on given axis"""
    if nbins is None:
        nbins = bin_edges[-1] - bin_edges[0]
    be = np.linspace(*bin_edges, nbins + 1)
    bin_centers = (be[1:] + be[:-1]) / 2
    total = np.zeros(len(bin_centers))
    found = np.zeros(len(bin_centers))

    data_remaining = data[[on_axis, 'acceptance_fraction']].copy()
    for bi in range(len(be) - 1):
        mask = (data_remaining[on_axis] >= be[bi]) & (data_remaining[on_axis] < be[bi + 1])
        total[bi] = len(data_remaining[mask])
        found[bi] = np.sum(data_remaining[mask]['acceptance_fraction'])
        data_remaining = data_remaining[~mask]

    values, yerr = get_interval(bin_centers, total, found)
    return bin_centers, values, yerr


def acceptance_plot(data, on_axis, bin_edges, nbins=None, plot_label=""):
    """
    Compute acceptance from data using acceptance_fraction
    (this is an arbitrary weighing of the acceptance based on the outcome of matching)
    """
    bin_centers, values, yerr = calc_arb_acceptance(data, on_axis, bin_edges, nbins)
    _plot_acc(bin_centers, values, yerr, plot_label)
    plt.xlabel(on_axis.replace('_', ' '))


def _plot_acc(bin_centers, values, yerr, plot_label):
    plt.errorbar(x=bin_centers,
                 y=values,
                 yerr=yerr,
                 linestyle='none',
                 marker='o',
                 markersize=4,
                 capsize=3,
                 label=plot_label,
                 )


def _plot_mh_percentile(mh: multihist.Histdd,
                        percentile: int,
                        **kwargs,
                        ):
    percentile_from_mh = mh.percentile(percentile, mh.axis_names[1])
    kwargs.setdefault('drawstyle', 'steps-mid', )
    plt.plot(percentile_from_mh.bin_centers,
             percentile_from_mh,
             **kwargs
             )


def rec_plot(dat,
             dpe_offset=0.0,
             show_hist=True,
             axis_names=('raw_area_trigger', 'rec_bias'),
             print_dpe=True,
             _percentiles_y_bins=500,
             _y_label_kwargs=immutabledict(fontsize=18),
             **kwargs):
    kwargs.setdefault('bins', 50)
    kwargs.setdefault('range', [[0, 50], [-1, 1]])
    x_ax, y_ax = axis_names

    mh_display = multihist.Histdd(axis_names=axis_names, **kwargs)

    # Make a second Multihist with the full y-range for reliable percentiles,
    # otherwise you might get wrong results since you cut away data from the
    # range kwargs.
    percentile_kwargs = deepcopy(kwargs)
    y_data = dat[y_ax] / (1 + dpe_offset) - 1
    percentile_kwargs['range'][1] = [y_data.min(), y_data.max()]
    if isinstance(kwargs['bins'], (list, tuple, np.ndarray)):
        percentile_kwargs['bins'] = [kwargs['bins'][0], _percentiles_y_bins]
    else:
        percentile_kwargs['bins'] = [kwargs['bins'], _percentiles_y_bins]
    mh_for_percentiles = multihist.Histdd(axis_names=axis_names, **percentile_kwargs)

    mh_display.add(dat[x_ax], y_data)
    mh_for_percentiles.add(dat[x_ax], y_data)

    one_sigma = 100 * (1 - norm.cdf(1))
    median = 50
    minus_one_sigma = 100 * (1 - norm.cdf(-1))
    _plot_mh_percentile(mh_for_percentiles, median,
                        color='whitesmoke', label='median')
    _plot_mh_percentile(mh_for_percentiles, one_sigma,
                        color='cyan', label=f'{one_sigma:.1f}% quantile')
    _plot_mh_percentile(mh_for_percentiles, minus_one_sigma,
                        color='green', label=f'{minus_one_sigma:.1f}% quantile')

    del mh_for_percentiles

    if show_hist:
        mh_display.plot(log_scale=True)
    a_rec = r'\frac{\mathrm{Area}_\mathrm{reconstructed}\mathrm{\ }[\mathrm{PE}]'
    dpe_div = r'\mathrm{\ }/\mathrm{\ }(1+p_{DPE})}'
    if print_dpe:
        dpe_div = dpe_div[:-1] + r'\mathrm{\ }|\mathrm{\ }p_{DPE}=' + f'{dpe_offset}' + '}'
    true_nph = DEFAULT_NPH_LABEL
    plt.xlabel(true_nph)
    y_label = r'$' + a_rec + dpe_div + r'{' + true_nph.replace('$', '') + r'}-1$'
    plt.ylabel(y_label, **_y_label_kwargs)
    plt.grid()


def _rec_kwargs(s1_kwargs=None,
                s2_kwargs=None):
    if s1_kwargs is None:
        s1_kwargs = {}
    if s2_kwargs is None:
        s2_kwargs = {}
    return s1_kwargs, s2_kwargs


def _rec_diff_inner(dat, title, **kwargs):
    if not len(dat):
        return
    rec_plot(dat, **kwargs)
    plt.axhline(0, linestyle='--', c='k')
    plt.title(title)


def reconstruction_bias(data, **kwargs):
    rec_diff(data, None, **kwargs)


def rec_diff(def_data,
             data_alt,
             data_set_names=("default", "custom"),
             s1_kwargs=immutabledict(dpe_offset=0.0),
             s2_kwargs=immutabledict(dpe_offset=0.0),
             ):
    make_diff = data_alt is not None
    data_sets = [def_data, data_alt] if make_diff else [def_data]

    f, axes = plt.subplots(2, len(data_sets), figsize=(18, 13))
    s1_kwargs, s2_kwargs = _rec_kwargs(s1_kwargs,
                                       s2_kwargs)
    axes = iter(axes.flatten())
    for axi, dat in enumerate(data_sets):
        plt.sca(next(axes))
        mask = (dat['type'] == 1) & (dat['outcome'] == 'found')
        _rec_diff_inner(dat[mask],
                        title=f'{data_set_names[axi]} S1 rec. bias',
                        **s1_kwargs)
        if axi == 0:
            plt.legend()
    for axi, dat in enumerate(data_sets):
        plt.sca(next(axes))
        mask = (dat['type'] == 2) & (dat['outcome'] == 'found')
        _rec_diff_inner(dat[mask],
                        title=f'{data_set_names[axi]} S2 rec. bias',
                        **s2_kwargs)
    return axes
